#|

A demonstration of a possibly peculiar (?) failure of CL+SSL in
threaded mode.  This glitch happens for both CCL and SBCL on a Mac.

To run this, ensure that ./server.crt and ./server.key exist (see
note on creating them) and run

lisp>  (test-https-server/good :port 8080)
 and run apache bench
shell>  ab  -f SSL2 -n 1000  -c 20  -v 1   https://127.0.0.1:8080

This version runs well with concurrency and threading, occasionally
dropping a connection at very high loads.

Then restart the lisp process and run the bad version:

lisp>  (test-https-server/bad :port 8080)
  and run it with no concurrency
shell>  ab  -f SSL2 -n 1000  -c 1  -v 1   https://127.0.0.1:8080
 This runs fine with concurrency=1
 Then try it with concurrency=10
shell>  ab  -f SSL2 -n 1000  -c 10  -v 1   https://127.0.0.1:8080
 And if fails with a mass of SSL errors.

The difference is that in the /good version the cl+ssl:make-ssl-server-stream
is run in the main acceptor thread, but in the /bad version
it is run inside the worker thread.

----

What does hunchentoot do?  If I understand it correctly, then
accept-connections calls 
 (handle-incoming-connection taskmaster client-connection)

The handle-incoming-connection method for the 
one-thread-per-connection-taskmaster
calls (create-request-handler-thread taskmaster socket).

Then create-request-handler-thread creates a thread
and within this thread 
  (process-connection (taskmaster-acceptor taskmaster) socket)) 
is called.

Then 

  (initialize-connection-stream *acceptor* 
       (make-socket-stream socket *acceptor*))
is run by process-connection.

In hunchentoot/ssl.lisp initialize-connection-stream 
is specialized on ssl-acceptor, and associates ssl with
the stream using l+ssl:make-ssl-server-stream

Hence it appears that hunchentoot is hit by the problem of 
creating an SSL within a child thread rather than the acceptor thread.


==========
NOTE - to make the certifcates just do the following, using 'pass' for password.
The commands have to be pasted in one by one and dummy values entered.


  openssl genrsa -des3 -out server.key 1024
  openssl req -new -key server.key -out server.csr
  openssl req -new -key server.key -out server.csr
  cp server.key server.key.org
  openssl rsa -in server.key.org -out server.key
  openssl x509 -req -days 365 -in server.csr -signkey server.key -out server.crt
==========
|#

(eval-when (load eval compile)
  (require 'cl+ssl)
  (require 'bordeaux-threads)
  (require 'usocket))



(defparameter *keyfile* "./server.key")
(defparameter *certfile* "./server.crt")

(defun read-line-crlf (stream &optional eof-error-p)
  (let ((s (make-string-output-stream)))
    (loop
        for empty = t then nil
        for c = (read-char stream eof-error-p nil)
        while (and c (not (eql c #\return)))
        do
          (unless (eql c #\newline)
            (write-char c s))
        finally
          (return
            (if empty nil (get-output-stream-string s))))))



;; the good server associates SSL with a stream inside the acceptor thread,
;; then passes this on to the worker thread
(defun test-https-server/good
    (&key (port 8080)
          (cert *certfile*)
          (key *keyfile*))
  (format t "~&SSL server listening on port ~d~%" port)
  (let ((server  (usocket:socket-listen "127.0.0.1" port 
					:backlog 100)))
    (loop
      (let* ((socket (usocket:socket-accept server))
	     ;; MAKE SSL HERE
             (client (cl+ssl:make-ssl-server-stream
                      (usocket:socket-stream socket)
                      :external-format '(:iso-8859-1 :eol-style :lf)
                      :certificate cert
                      :key key)))
	;; launch response in a separate thread
	(bordeaux-threads:make-thread
	   (lambda ()
	     (unwind-protect
		  (progn
		    (format client "HTTP/1.0 200 OK~%")
		    (format client "Server: SSL-TEST/1.1~%")
		    (format client "Content-Type: text/plain~%")
		    (terpri client)
		    (format client "HELLO WORLD")
		    (finish-output client))
	     (close client))))))))


;; the bad server associates an SSL with the client inside a 
;; worker thread, not inside the main thread
(defun test-https-server/bad
    (&key (port 8080)
          (cert *certfile*)
          (key *keyfile*))
  (format t "~&SSL server listening on port ~d~%" port)
  (let ((server  (usocket:socket-listen "127.0.0.1" port 
					:backlog 100)))
    (loop
      (let* ((socket (usocket:socket-accept server))
	     (stream (usocket:socket-stream socket)))
	;; launch response in a separate thread
	(bordeaux-threads:make-thread
	   (lambda ()
	     ;; MAKE SSL HERE
	     (let ((client (cl+ssl:make-ssl-server-stream
			    stream
			    :external-format '(:iso-8859-1 :eol-style :lf)
			    :certificate cert
			    :key key)))
	       (unwind-protect
		    (progn
		      (format client "HTTP/1.0 200 OK~%")
		      (format client "Server: SSL-TEST/1.1~%")
		      (format client "Content-Type: text/plain~%")
		      (terpri client)
		      (format client "HELLO WORLD")
		      (finish-output client))
		 (close client)))))))))


